module mult_nmo_mpi

use mpi_sep

implicit none

logical, private, save :: verbose

contains

! ---------------------------------------------------------------------------
subroutine mult_nmo_mpi_init( verb)
	logical, intent(IN) :: verb

	if(verb) write(0,*) 'mult_nmo_mpi_init(): TOP'
	verbose = verb
	if(verbose) write(0,*) 'mult_nmo_mpi_init(): BOT'
end subroutine mult_nmo_mpi_init
! ---------------------------------------------------------------------------

! ---------------------------------------------------------------------------
! If baselabel="vel" and ithr=2, outlabel="vel_2"
! ---------------------------------------------------------------------------
subroutine mpi_distrib_string( baselabel, ithr, outlabel, datapath )
	integer,           intent(IN)   :: ithr
	character(len=20), intent(IN)   :: baselabel
	character(len=200), intent(OUT) :: outlabel
	character(len=10)               :: tag_num
	character(len=180)              :: scrpath, machine
	character(len=200)              :: datapath
	integer                         :: ilen, ierr
	optional                        :: datapath

#ifdef SEP_MPI
	call MPI_Get_processor_name(machine,ilen,ierr)
	write( scrpath, "(a,a,a)" ) "/net/",machine(1:ilen),"/scr1/temp/"
#else
	scrpath="./"	
#endif

	if(present(datapath)) write(datapath,"(a,a)") "datapath=",scrpath

	if( ithr < 10) then
		write(tag_num,"(a,i1)") "_tmp_",ithr
	else if( ithr < 100 ) then
		write(tag_num,"(a,i2)") "_tmp_",ithr
	else
		write(tag_num,"(a,i3)") "_tmp_",ithr
	end if

	write(outlabel,"(a,a,a)") trim(scrpath),trim(baselabel),trim(tag_num)

end subroutine mpi_distrib_string
! ---------------------------------------------------------------------------

! ---------------------------------------------------------------------------
! Broadcast an auxiliary file to all nodes.  Only called if MPI_SEP defined
! ---------------------------------------------------------------------------
subroutine mpi_aux_read_checkin( baselabel, ithr, ndata )
	character(len=20), intent(IN)  :: baselabel
	character(len=200)             :: thislabel, datapath
	integer,           intent(IN)  :: ithr, ndata
	integer                        :: ierr
	integer, external              :: mpi_sep_tag_bcast, auxinout, sseek
	character(len=2)               :: tmp_str

	if(verbose) write(0,*) 'mpi_aux_read_checkin(): TOP'

	call mpi_distrib_string( baselabel, ithr, thislabel, datapath=datapath)
	ierr = auxinout(thislabel)
	call getch_add_string(datapath)
	ierr = mpi_sep_tag_bcast( 0, baselabel, ithr, thislabel, ndata, 1)
	if( ierr /= 0) then
		write(0,*) 'mpi_aux_read_checkin(): Error broadcasting to file ',thislabel
		call exit(0)
	end if
	ierr = sseek(thislabel,0,0)

	if(verbose) write(0,*) 'mpi_aux_read_checkin(): BOT'
end subroutine mpi_aux_read_checkin
! ---------------------------------------------------------------------------

! ---------------------------------------------------------------------------
! Checkout aux file after broadcast+distrib.  Only called if MPI_SEP defined
! ---------------------------------------------------------------------------
subroutine mpi_aux_read_checkout( baselabel, ithr, ndata, nbytes, test )
	character(len=20), intent(IN) :: baselabel
	character(len=200)            :: thislabel
	integer,           intent(IN) :: ndata, nbytes, ithr
	logical,           intent(IN) :: test
	integer, external             :: sseek
	integer                       :: stat

	if(verbose) write(0,*) 'mpi_aux_read_checkout(): TOP'

	call mpi_distrib_string( baselabel, ithr, thislabel)
	if( nbytes /= ndata*4 ) then
		write(0,"(a,a,a,i10,i10)") 'mpi_aux_read_checkout(): Error reading from file ',&
		                            trim(thislabel),': nbytes /= ndata*4',nbytes,ndata*4
		call exit(0)
	else
		if(test) write(0,"(a,i2,a,i10,a,a)") 'thread ',ithr,': mpi_aux_read_checkout(): read ',&
		                                     nbytes,' bytes from file ',thislabel
	end if
	stat = sseek( thislabel, 0, 0)

	if(verbose) write(0,*) 'mpi_aux_read_checkout(): BOT'
end subroutine mpi_aux_read_checkout
! ---------------------------------------------------------------------------

! ---------------------------------------------------------------------------
! Broadcast a blocked file to thread "ithr".  Only called if MPI_SEP defined
! ---------------------------------------------------------------------------
subroutine mpi_distrib_read_checkin( baselabel, ithr, ndata, nblock, isend )
	character(len=20),     intent(IN)  :: baselabel
	character(len=200)                 :: thislabel, datapath
	integer,               intent(IN)  :: ithr, ndata, nblock
	integer, dimension(:), intent(IN)  :: isend
	integer                            :: ierr, stat
	integer, external                  :: mpi_sep_tag_distribute, auxinout, sseek

	if(verbose) write(0,*) 'mpi_distrib_read_checkin(): TOP'

	call mpi_distrib_string( baselabel, ithr, thislabel, datapath=datapath)
	ierr = auxinout(thislabel)
	call getch_add_string(datapath)
	ierr = mpi_sep_tag_distribute( 0, baselabel, ithr, thislabel, ndata, nblock, isend)
	if( ierr/=0 ) then
		write(0,"(a,a,a,i2)") 'mpi_distrib_read_checkin(): Error distributing to file ',&
		                       thislabel,': ithr=',ithr
		call exit(0)
	end if
	stat = sseek( thislabel, 0, 0)

	if(verbose) write(0,*) 'mpi_distrib_read_checkin(): BOT'

end subroutine mpi_distrib_read_checkin
! ---------------------------------------------------------------------------

! ---------------------------------------------------------------------------
subroutine mpi_distrib_read_checkout( baselabel, ithr, ndata, nbytes, test )
	character(len=20), intent(IN) :: baselabel
	character(len=200)             :: thislabel
	integer,           intent(IN) :: ndata, nbytes, ithr
	logical,           intent(IN) :: test
	integer, external             :: sseek
	integer                       :: stat

	if(verbose) write(0,*) 'mpi_distrib_read_checkout(): TOP'

	call mpi_distrib_string( baselabel, ithr, thislabel)
	if( nbytes /= ndata*4 ) then
		write(0,"(a,a,a,i10,i10)") 'mpi_distrib_read_checkout(): Error reading from file ',&
		                            trim(thislabel),': nbytes /= ndata*4',nbytes,ndata*4
		call exit(0)
	else
		if(test) write(0,"(a,i2,a,i10,a,a)") 'thread ',ithr,&
		                                     ': mpi_distrib_read_checkout(): read ',&
		                                      nbytes,' bytes from file',thislabel
	end if
	stat = sseek( thislabel, 0, 0)

	if(verbose) write(0,*) 'mpi_distrib_read_checkout(): BOT'
end subroutine mpi_distrib_read_checkout
! ---------------------------------------------------------------------------

!--------------------------------------------------------------------------------
! Collect blocked file from nodes and collocate to node 0.
!--------------------------------------------------------------------------------
subroutine mpi_distrib_collect( baselabel, ithr, ndata1, ndata2, nblock,        &
                                nbinner, isend, nbytes, test                    )

	character(len=20),     intent(IN) :: baselabel
	character(len=200)                :: thislabel
	integer,               intent(IN) :: ithr, ndata1, ndata2, nblock, nbytes,    &
	                                     nbinner
	logical,               intent(IN) :: test
	integer, dimension(:), intent(IN) :: isend
	integer                           :: stat
	integer, external                 :: mpi_sep_tag_combine_big, sseek

	if(verbose) write(0,*) 'mpi_distrib_collect(): TOP'

	call mpi_distrib_string( baselabel, ithr, thislabel)
	if( nbytes /= ndata1*ndata2*nbinner*4 ) then
		write(0,*) 'mpi_distrib_collect(): Error writing from file ',thislabel
		call exit(0)
	else
		if(test) write(0,"(a,i10,a,a)") 'mpi_distrib_collect(): wrote ',nbytes,&
		                                ' bytes from file ',thislabel
	end if
#ifdef SEP_MPI
  call auxclose(thislabel)
  call auxin(thislabel)
	if( 0 /= mpi_sep_tag_combine_big( ithr, thislabel, 0, baselabel, ndata1,  &
	                                  ndata2, nblock, isend, 0) ) then
		write(0,*) 'mpi_distrib_collect(): Error from file tag ',thislabel
		call exit(0)
	end if
#endif

  call auxclose(thislabel)

	if(verbose) write(0,*) 'mpi_distrib_collect(): BOT'

end subroutine mpi_distrib_collect
! ---------------------------------------------------------------------------

! ---------------------------------------------------------------------------
! Reads and distributes auxfile, with optional size checking.  Calls internal
! subroutines mpi_aux_read_checkin() and mpi_aux_read_checkout().
! ---------------------------------------------------------------------------
subroutine read_dist_aux( baselabel, thislabel, data, ithr, ndata, test,    &
                          check1, check2, check3, default                   )
	character(len=*),   intent(IN)    :: baselabel, thislabel
	real, dimension(:), intent(INOUT) :: data
	integer,            intent(IN)    :: ithr, ndata
	integer                           :: check1, check2, check3, ntmp, nbytes
	character(len=20)                 :: baselab20
	character(len=200)                :: thislab20
	character(len=100)                :: file
	character(len=300)                :: errstr
	real                              :: default
	logical                           :: test
	optional                          :: test, check1, check2, check3, default
	integer, external                 :: sreed, getch

	if(verbose) write(0,*) 'read_dist_aux(): TOP'

	if( getch(baselabel,'s',file) /= 0 ) then            ! Test file existence


		baselab20=baselabel;  thislab20=thislabel
		if(present(check1)) then
			call from_aux( baselabel, "n1", ntmp)
			write(errstr,*) 'ERROR: ',thislabel,'(n1) !=',baselabel,'(n1)'
			if( ntmp /= check1) call seperr( trim(errstr)) 
		end if
		if(present(check2)) then
			call from_aux( baselabel, "n2", ntmp)
			write(errstr,*) 'ERROR: ',thislabel,'(n2) !=',baselabel,'(n2)'
			if( ntmp /= check2) call seperr( trim(errstr)) 
		end if
		if(present(check3)) then
			call from_aux( baselabel, "n3", ntmp)
			write(errstr,*) 'ERROR: ',thislabel,'(n3) !=',baselabel,'(n3)'
			if( ntmp /= check3) call seperr( trim(errstr)) 
		end if

#ifdef SEP_MPI
		call mpi_aux_read_checkin( baselab20, ithr, ndata)
		nbytes = sreed( thislab20, data, ndata*4)
		call mpi_aux_read_checkout( baselab20, ithr, ndata, nbytes, test)
#else
		nbytes = sreed( baselabel, data, ndata*4)
		if(test) write(0,"(a,i10,a,a)") 'read_dist_aux(): read',&
		                                 nbytes,'from file',thislabel
#endif
	else
		if(verbose) write(0,"(a,a,a,f)") 'Setting file ',baselabel,' to default: ',default
		if( present(default) ) data = default
	end if

	if(verbose) write(0,*) 'read_dist_aux(): BOT'

end subroutine read_dist_aux
! ---------------------------------------------------------------------------

! ---------------------------------------------------------------------------
! Reads and distributes blocked file. 
! ---------------------------------------------------------------------------
subroutine read_dist_main( baselabel, thislabel, data, ithr, ndata, nblock, &
                           ninner, isend, test                              )
	character(len=*),   intent(IN)    :: baselabel, thislabel
	real, dimension(:), intent(INOUT) :: data
	integer,            intent(IN)    :: ithr, ndata, nblock, ninner, isend(:)
	integer                           :: nbytes
	character(len=200)                :: thislab20
	character(len=20)                 :: baselab20
	logical                           :: test
	optional                          :: test
	integer, external                 :: sreed
	
	if(verbose) write(0,*) 'read_dist_main(): TOP'

	baselab20=baselabel;  thislab20=thislabel

#ifdef SEP_MPI
	call mpi_distrib_read_checkin( baselab20, ithr, ndata, nblock, isend)
	nbytes = sreed( thislab20, data, ninner*ndata*4)
	call mpi_distrib_read_checkout( baselab20, ithr, ninner*ndata, nbytes, test)
#else
	nbytes = sreed( baselabel, data, nblock*ndata*4)
	if(test) write(0,"(a,i10,a,a)") 'read_dist_main(): read',&
	                                 nbytes,'from file',thislabel
#endif

	if(verbose) write(0,*) 'read_dist_main(): bot'

end subroutine read_dist_main
! ---------------------------------------------------------------------------

!--------------------------------------------------------------------------------
! Collects and writes blocked file. 
!--------------------------------------------------------------------------------
subroutine collect_write_main( baselabel, thislabel, data, ithr, ndata1,        &
                               ndata2, nblock, ninner, isend, test              )

	character(len=*),   intent(IN)  :: baselabel, thislabel
	real, dimension(:), intent(IN)  :: data
	integer,            intent(IN)  :: ithr, ndata1, ndata2, nblock, ninner,      &
	                                   isend(:)
	integer                         :: nbytes, ierr
	character(len=200)              :: thislab20
	character(len=20)               :: baselab20
	logical                         :: test
	optional                        :: test
	integer, external               :: srite
	
	if(verbose) write(0,*) 'collect_write_main(): TOP'

	baselab20=baselabel;  thislab20=thislabel

#ifdef SEP_MPI
	nbytes = srite( thislab20, data, ninner*ndata1*ndata2*4)
	call mpi_distrib_collect( baselab20, ithr, ndata1, ndata2, nblock, ninner,    &
	                          isend, nbytes, test                                 )
#else
	nbytes = srite( baselabel, data, ninner*ndata1*ndata2*4)
	if(test) write(0,"(a,i10,a,a)") 'collect_write_main(): wrote',&
	                                 nbytes,'to file ',baselabel
#endif

	if(verbose) write(0,*) 'collect_write_main(): bot'

end subroutine collect_write_main
!--------------------------------------------------------------------------------

end module mult_nmo_mpi
