module mult_nmo_reg_3d

use adj_mod
use chain0_mod
use mult_nmo_util
use morgan_util
use dottest
use scaled_identity

implicit none

integer, private, save  :: nt, nx1, nx2, nmod, ithr, ncmp, nsurf
real,    private, save  :: eps_grad, eps_order, eps_xtalk
logical, private, save  :: verbose

real, dimension(:), private, pointer, save        :: tmp1, tmp2, tmp3, tmpwt 
real, dimension(:), private, allocatable, save    :: xtalkwt
integer, dimension(:), private, allocatable, save :: nullcube

logical, private, parameter, save :: T = .true., F = .false.

contains

!-------------------------------------------------------------------------------
! Initialization
!-------------------------------------------------------------------------------
	subroutine mult_nmo_reg_init( nt_in, nx1_in, nx2_in, maxorder, nsurf_in,     &
	                              eps_order_in, eps_grad_in, eps_xtalk_in, verb, &
	                              ithread                                        )

		real,    intent(IN) :: eps_order_in, eps_grad_in, eps_xtalk_in
		integer, intent(IN) :: nt_in, nx1_in, nx2_in, maxorder, nsurf_in
		integer             :: ithread
		logical             :: verb
		optional            :: verb, ithread

		if(verb) write(0,*) 'mult_nmo_reg_init(): TOP'

		nt = nt_in;  nx1 = nx1_in;  nx2 = nx2_in
		ncmp = nt*nx1*nx2
		nsurf = nsurf_in
		nmod = get_nmod(maxorder,nsurf)

		eps_order = eps_order_in
		eps_grad  = eps_grad_in
		eps_xtalk = eps_xtalk_in
		verbose = .false. ; if( present(verb)     ) verbose = verb
		ithr    = 1       ; if( present(ithread)  ) ithr = ithread

		allocate( tmp1(ncmp*nmod*2), tmp2(ncmp*nmod), tmp3(ncmp*nmod),             &
		          tmpwt(2*ncmp*nmod), xtalkwt(ncmp*nmod), nullcube(ncmp*nmod)      )

		nullcube = 1

		if(verbose) write(0,*) 'mult_nmo_reg_init(): BOT'
	end subroutine mult_nmo_reg_init
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! Initialize crosstalk weights for this CMP gather.
!-------------------------------------------------------------------------------
subroutine mnr_set_xtalk( crosstalk, nullcube_in)
	real, dimension(nt*nx1*nx2*nmod) :: crosstalk
	integer, dimension(nt*nx1*nx2*nmod) :: nullcube_in
	if(verbose) write(0,*) 'mnr_set_xtalk(): TOP'

  xtalkwt = crosstalk
  nullcube = nullcube_in

	if(verbose) write(0,*) 'mnr_set_xtalk(): BOT'
end subroutine mnr_set_xtalk
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
! Clean up after each CMP gather
!-------------------------------------------------------------------------------
	subroutine mult_nmo_reg_close( )
		if(verbose) write(0,*) 'mult_nmo_reg_close(): TOP'
		if( .NOT. allocated(xtalkwt) ) write(0,*) 'WARNING: xtalkwt not allocated'
		if( .NOT. allocated(tmp1)    ) write(0,*) 'WARNING: tmp1 not allocated'
		if( .NOT. allocated(tmp2)    ) write(0,*) 'WARNING: tmp2 not allocated'
		if( .NOT. allocated(tmp3)    ) write(0,*) 'WARNING: tmp3 not allocated'
		if( .NOT. allocated(tmpwt)   ) write(0,*) 'WARNING: tmpwt not allocated'
		deallocate( xtalkwt, tmp1, tmp2, tmp3, tmpwt)
		if(verbose) write(0,*) 'mult_nmo_reg_close(): BOT'
	end subroutine mult_nmo_reg_close
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! Weighted offset gradient
!-------------------------------------------------------------------------------
integer function nmo_row_igrad_exec( adj, add, model, data)
	logical, intent(IN) :: adj, add
	real, dimension(:)  :: model, data

	call adjnull( adj, add, model, data)
	if(verbose) call oper_stats( 'nmo_row_igrad_exec()', .true., model, data,    &
	                             adj, ithr=ithr                                  )

	call scaled_identity_init( eps_grad)
	call chain0( scaled_identity_exec, nmo_row_igrad_core, adj, add, model,      &
	             data, tmpwt                                                     )

	if(verbose) call oper_stats( 'nmo_row_igrad_exec()', .false., model, data,   &
	                             adj, ithr=ithr                                  )
	nmo_row_igrad_exec = 0
end function nmo_row_igrad_exec
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
! 2-D gradient across offset of all model panels 
!-------------------------------------------------------------------------------
integer function nmo_row_igrad_core( adj, add, model, data)
	integer             :: i1, i2, i3, j1, j2, it, ix1, ix2, imod, k1, k2
	integer             :: nd4(4), nm5(5), im4(4)
	logical, intent(IN) :: adj, add
	real, dimension(:)  :: model, data

	call adjnull( adj, add, model, data)
	if(verbose) call oper_stats( 'nmo_row_igrad_core()', .true., model, data,    &
	                             adj, ithr=ithr                                  )
	if( eps_grad/=0.0 ) then

		nd4 = (/nt,nx1,nx2,nmod/)
		nm5 = (/nt,nx1,nx2,nmod,2/)

		do imod=1,nmod
			do it=1,nt
				do ix2=1,max(1,nx2-1)
					k2 = min(ix2+1,nx2)
					do ix1=1,max(1,nx1-1)         ! Unrealistic (nx1=1) but safe
						k1 = min(ix1+1,nx1)
						im4 = (/it,ix1,ix2,imod/)
						i1 = grid2line4( nd4, im4 );  if( nullcube(i1)/=1 ) cycle
						im4 = (/it,k1,ix2,imod/)
						i2 = grid2line4( nd4, im4 );  if( nullcube(i2)/=1 ) cycle
						im4 = (/it,ix1,k2,imod/)
						i3 = grid2line4( nd4, im4 );  if( nullcube(i3)/=1 ) cycle
						j1 = grid2line5( nm5, (/it,ix1,ix2,imod,1/) )
						j2 = grid2line5( nm5, (/it,ix1,ix2,imod,2/) )
						if(adj) then
							model(i1) = model(i1) - data(j1)
							model(i1) = model(i1) - data(j2)
							model(i2) = model(i2) + data(j1)
							model(i3) = model(i3) + data(j2)
						else
							data(j1) = data(j1) + ( model(i2) - model(i1) )
							data(j2) = data(j2) + ( model(i3) - model(i1) )
						end if
					end do
				end do
			end do
		end do

	end if

	if(verbose) call oper_stats( 'nmo_row_igrad_core()', .false., model, data,   &
	                             adj, ithr=ithr                                  )
	nmo_row_igrad_core = 0
end function nmo_row_igrad_core
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! Weighted order gradient
!-------------------------------------------------------------------------------
integer function nmo_order_igrad_exec( adj, add, model, data)
	logical, intent(IN) :: adj, add
	real, dimension(:)  :: model, data

	call adjnull( adj, add, model, data)
	if(verbose) call oper_stats( 'nmo_order_igrad_exec()', .true., model, data,  &
	                             adj, ithr=ithr                                  )

	call scaled_identity_init( eps_order)
	call chain0( scaled_identity_exec, nmo_order_igrad_core, adj, add, model,    &
	             data, tmpwt(1:ncmp*nmod)                                        )

	if(verbose) call oper_stats( 'nmo_order_igrad_exec()', .false., model, data, &
	                             adj, ithr=ithr                                  )
	nmo_order_igrad_exec = 0
end function nmo_order_igrad_exec
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! 1-D gradient across order (fixed t,x).
!-------------------------------------------------------------------------------
integer function nmo_order_igrad_core( adj, add, model, data)
	logical, intent(IN) :: adj, add
	real, dimension(:)  :: model, data
	integer             :: ilo, ihi, it, ix1, ix2, iorder, k, nd4(4), im4(4)

	call adjnull( adj, add, model, data)

	if(verbose) call oper_stats( 'nmo_order_igrad_core()', .true., model, data,  &
	                             adj, ithr=ithr                                  )
	if( eps_order /= 0.0) then

		nd4 = (/nt,nx1,nx2,nmod/)

		do iorder=1,max(1,nmod-1)
			k = min(nmod,iorder+1)
			do it=1,nt
				do ix2=1,nx2
					do ix1=1,nx1
						im4 = (/it,ix1,ix2,iorder/)
						ilo = grid2line4( nd4, im4 );  if( nullcube(ilo)/=1 ) cycle
						im4 = (/it,ix1,ix2,k/)
						ihi = grid2line4( nd4, im4 );  if( nullcube(ihi)/=1 ) cycle
						if(adj) then
							model(ilo) = model(ilo) - data(ilo)
							model(ihi) = model(ihi) + data(ilo)
						else
							data(ilo) = data(ilo) + ( model(ihi) - model(ilo) )
						end if
					end do
				end do
			end do
		end do

	end if

	if(verbose) call oper_stats( 'nmo_order_igrad_core()', .false., model, data, &
	                             adj, ithr=ithr                                  )
	nmo_order_igrad_core = 0
end function nmo_order_igrad_core
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! Apply scaled offset and order gradients as a vertical array operator.
!-------------------------------------------------------------------------------
integer function nmo_gradchain_exec( adj, add, model, data)
	integer              :: nn
	logical, intent(IN)  :: adj, add
	real, dimension(:)   :: model, data

	call adjnull( adj, add, model, data)
	if(verbose) call oper_stats( 'nmo_gradchain_exec()', .true., model, data,    &
	                             adj, ithr=ithr                                  )
	nn = size(data)/4

	tmp1 = data(1:2*nn)                                 ! For 2-D offset gradient
	tmp2 = data(2*nn+1:3*nn)                            ! For 1-D order gradient
	tmp3 = data(3*nn+1:4*nn)                            ! For crosstalk weighting

	call array0( nmo_row_igrad_exec, nmo_order_igrad_exec, noff_wt_exec, &
               adj, add, model, tmp1, tmp2, tmp3 )

	data(1:2*nn)      = tmp1
	data(2*nn+1:3*nn) = tmp2
	data(3*nn+1:4*nn) = tmp3

	if(verbose) call oper_stats( 'nmo_gradchain_exec()', .false., model, data,   &
	                             adj, ithr=ithr                                  )
	nmo_gradchain_exec = 0
end function nmo_gradchain_exec
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! Apply (weighted) crosstalk weight
!-------------------------------------------------------------------------------
integer function noff_wt_exec( adj, add, model, data)
	integer             :: im0, im1, iorder, stat
	logical, intent(IN) :: adj, add
	real, dimension(:)  :: model, data

	call adjnull( adj, add, model, data)
	if(verbose) call oper_stats( 'noff_wt_exec()', .true., model, data, adj,     &
	                             ithr=ithr                                       )
	if( eps_xtalk/=0.0 ) then

		if( adj) then
			model = model + data*xtalkwt*eps_xtalk
		else
			data = data + model*xtalkwt*eps_xtalk
		end if

	end if

	if(verbose) call oper_stats( 'noff_wt_exec()', .false., model, data, adj,    &
	                             ithr=ithr                                       )
	noff_wt_exec = 0
end function noff_wt_exec
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! Dot product test for all operators in this module
!-------------------------------------------------------------------------------
subroutine mnr_dp_test()
	real :: dot1(2), dot2(2)

	write(0,*) '----- MULT_NMO_REG MODULE DOT-PRODUCT TESTS ------'
	write(0,*) '----- each row should be have same numbers -------'
	write(0,*) '----- Testing nmo_row_igrad_exec -----'
	dot1=0.0;  dot2=0.0
	call dot_test( nmo_row_igrad_exec, nmod*nt*nx1*nx2, nmod*nt*nx1*nx2*2, dot1, dot2)
	write(0,*) '----- Testing nmo_order_igrad_exec -----'
	dot1=0.0;  dot2=0.0
	call dot_test( nmo_order_igrad_exec, nmod*nt*nx1*nx2, nmod*nt*nx1*nx2, dot1, dot2)
	write(0,*) '----- Testing nmo_gradchain_exec -----'
	dot1=0.0;  dot2=0.0
	call dot_test( nmo_gradchain_exec, nmod*nt*nx1*nx2, 4*nmod*nt*nx1*nx2, dot1, dot2)
	write(0,*) '----- Testing noff_wt_exec -----'
	dot1=0.0;  dot2=0.0
	call dot_test( noff_wt_exec, nmod*nt*nx1*nx2, nmod*nt*nx1*nx2, dot1, dot2)
	write(0,*) '--------------------------------------------------'
end subroutine mnr_dp_test
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
subroutine mult_nmo_reg_3d_test()
	integer :: nn, stat, ix, it, imod, j0, j1
	integer, dimension(:), pointer :: nullcube_tmp
	real,    dimension(:), pointer :: crosstalk_tmp, d, m

	call mult_nmo_reg_init( 100, 20, 1, 1, 2, 1.0, 1.0, 1.0, F)
	nn = ncmp*nmod
	allocate( crosstalk_tmp(nn), nullcube_tmp(nn), d(2*nn), m(nn) )
	crosstalk_tmp = 2.0
	nullcube_tmp = 1.0
	call mnr_set_xtalk( crosstalk_tmp, nullcube_tmp)

	call mnr_dp_test()

	m = 1.0
	write(0,*) '-------- BEGIN FEASIBILITY TESTING ---------'
	write(0,*) '-------- TESTING NMO_ROW_IGRAD_EXEC --------'
	write(0,*) '-- all should be one --'
	do ix=1,20
		do it=1,100
			do imod=1,5
				j0 = grid2line3( (/100,20,5/), (/it,ix,imod/) )
				m(j0) = ix
			end do
		end do
	end do
	stat = nmo_row_igrad_exec( F, F, m, d)
	write(0,*) '-- Fixed it=50, imod=3, variable x. should be 1 --'
	do ix=1,20
		j0 = grid2line3( (/100,20,5/), (/50,ix,3/) )
		write(0,*) 'ix,diff',ix,d(j0)
	end do
	write(0,*) '------- TESTING NMO_ORDER_IGRAD_EXEC -------'
	write(0,*) '-- all should be one --'
	do imod=1,5
		do ix=1,20
			do it=1,100
				j0 = grid2line3( (/100,20,5/), (/it,ix,imod/) )
				m(j0) = imod
			end do
		end do
	end do
	stat = nmo_order_igrad_exec( F, F, m, d(1:nn))
	write(0,*) '-- Fixed it=50, ix=10, variable mod. should be 1 --'
	do imod=1,5
		j0 = grid2line3( (/100,20,5/), (/50,10,imod/) )
		write(0,*) 'imod,diff',imod,d(j0)
	end do
	write(0,*) '----------- TESTING NOFF_WT_EXEC -----------'
	write(0,*) '-- all should be 2.0 --'
	stat = noff_wt_exec( F, F, m, d(1:nn))
	write(0,*) 'min/max:',minval(d(1:nn)),maxval(d(1:nn))

	call mult_nmo_reg_close()
	deallocate( crosstalk_tmp, nullcube_tmp, d, m)
end subroutine mult_nmo_reg_3d_test
!-------------------------------------------------------------------------------

end module mult_nmo_reg_3d
